/*
 Instructors: Marco
 */
#include <iostream>
#include <sstream>
#include <gecode/driver.hh>
#include <gecode/int.hh>
#include <gecode/minimodel.hh>

#include "instance.h"

using namespace Gecode;

class MoreOptions: public Options {
public:
	/// Name of the directory where the fiels instructors.csv and courses.csv are
	std::string dirname;
	/// Initialize options with file name \a s
	MoreOptions(const char* s) :
			Options(s) {
	}
	/// Parse options from arguments \a argv (number is \a argc)
	void parse(int& argc, char* argv[]) {
		// Parse regular options
		Options::parse(argc, argv);
		// Dirname, should be at position 1
		if (argc == 1) {
			help();
			exit(1);
		}
		dirname = argv[1];
		argc--;
	}
	/// Print help message
	virtual void help(void) {
		Options::help();
		std::cerr << "\t(string) " << std::endl << "\t\t directory to pass"
				<< std::endl;
	}
};

class CIA: public IntMinimizeScript {
protected:

	const FlatInstance *instance;
	const int ni; // number of instructors
	const int nc; // number of courses

	IntVarArray instructor; // the instructor assigned to the course
	IntVarArray course; // the course assigned to the instructor
	IntVar nr_unassigned;
	IntVar worstPref;
	IntVar globalSat;
public:

	CIA(const MoreOptions& opt) :
			instance(new FlatInstance(opt.dirname.c_str())), ni(instance->ni), nc(
					instance->nc) {

		const int rangesI[][2] = { { 0, ni - 1 }, { ni, ni } };
		IntSet a(rangesI, 2);
		instructor = IntVarArray(*this, nc + 1, a);
		const int rangesC[2][2] = { { 0, nc - 1 }, { nc, nc } };
		IntSet b(rangesC, 2);
		course = IntVarArray(*this, ni + 1, b); //IntSet(rangesC, 2));
		nr_unassigned = IntVar(*this, 0, nc);
		worstPref = IntVar(*this, 1, ni);
		globalSat = IntVar(*this, 0, Int::Limits::max);
		// instance->printData(); // to print input data
		// instance->printRankings(); // to print rankings
		// The rank or value of the instructor ni-1 for course 0
		// std::cout << instance->instrRanking[0][ni - 1] << std::endl;
		// The rank or value of the course nc-1 for instructor 0
		// std::cout << instance->crsRanking[0][nc - 1] << std::endl;

		// Instantiate ================================================================
		string fileinstance = "./preassigned.txt";
		ifstream in(fileinstance.c_str());
		if (!in) {
			cerr << "Couldn't open the file " << fileinstance << endl;
			exit(1);
		}
		cout << "Read " << fileinstance << endl;
		std::string line;
		int skip = 0, preassigned = 0;
		DictionaryType crs_name_to_id = instance->crs_name_to_id;
		DictionaryType instr_name_to_id = instance->instr_name_to_id;
		while (std::getline(in, line)) {
			if (skip++ < 5)
				continue;
			std::istringstream iss(line);
			string a, b;
			if (!(iss >> a >> b)) {
				cerr << "Something wrong reading +" << iss.str() << "+"
						<< " on line " << 5 + preassigned << endl;
				break;
			} // error
			  //vector<int> crsids = crs2IDs(a);
			  //vector<int> instrids = instr2IDs(b);
			DictionaryType::iterator ita = crs_name_to_id.find(a);
			DictionaryType::iterator itb = instr_name_to_id.find(b);
			rel(*this, instructor[ita->second] == itb->second);
			crs_name_to_id.erase(ita);
			instr_name_to_id.erase(itb);
			preassigned++;
		}
		std::ostringstream s;
		s << "Post instantiation of " << preassigned << " courses... ";
		checkConsistency(s.str());
		// ===============================================================================

		// INSERT YOUR MODEL HERE
		// linking: the instructor of a course must be the course instructor
		// instructor[course[i]]=i;
		// if the instructor has a course then the course's instructor is i
		for (int i = 0; i < ni; i++)
			rel(*this,
					(course[i] == nc) || (element(instructor, course[i]) == i));
		// linking: the course of an instructor must be the course instructor
		for (int c = 0; c < nc; c++)
			rel(*this,
					(instructor[c] == ni)
							|| (element(course, instructor[c]) == c));
		checkConsistency("Post linking...");
		// ----------------------------------------------------------------------------
		// the instantiated rank of the course assinged to corresponding instructor
		IntVarArray insRankCourse(*this, ni, 0, nc);
		for (int i = 0; i < ni; i++) {
			IntArgs rankCourses;
			const int *tmp = instance->CrsRanking(i);
			rankCourses << IntArgs(nc, instance->CrsRanking(i))
					<< IntArgs(1, nc);
			//rel(*this, element(rankCourses, course[i]) == rankCourse_instanz);
			element(*this, rankCourses, course[i], insRankCourse[i], opt.icl());
			//rel(*this,element(rankCourses, course[i]) == rankCourse_res, opt.icl());
		}
		// the instantiated rank of the instructor assigned to corresponding course
		IntVarArray insRankInstructor(*this, nc, 0, ni);
		for (int c = 0; c < nc; c++) {
			IntArgs rankInstructors;
			rankInstructors << IntArgs(ni, instance->InstrRanking(c))
					<< IntArgs(1, ni);
			element(*this, rankInstructors, instructor[c], insRankInstructor[c],
					opt.icl());
		}
		// ----------------------------------------------------------------------------
		// constraint on preferences
		//forall(i in Course, i in Instructors)
		//crsRanking[i][c] < csrRanking[i][course[i]] => instrRanking[c][instructor[c]] <= instrRanking[c][i]);
		for (int i = 0; i < ni; i++) {
			BoolVar b1 = expr(*this, course[i] < nc);
			rel(*this, (b1) >> (insRankCourse[i] < worstPref));
			for (int c = 0; c < nc; c++) {
				rel(*this,
						((b1) && (instance->CrsRanking(i, c) < insRankCourse[i]))
								>> (insRankInstructor[c]
										<= instance->InstrRanking(c, i)));
			}
		}
		checkConsistency("Post stability 1... ");
		//instrRanking[c][i] < instrRanking[instructor[c]][i] => crsRanking[i][course[i]] <= crsRanking[i][c]);
		for (int c = 0; c < nc; c++) {
			BoolVar b2 = expr(*this, instructor[c] < ni);
			// Worst preference used:
			rel(*this, (b2) >> (insRankInstructor[c] < worstPref));
			for (int i = 0; i < ni; i++) {
				rel(*this,
						((b2)
								&& (instance->InstrRanking(c, i)
										< insRankInstructor[c]))
								>> (insRankCourse[i]
										<= instance->CrsRanking(i, c)));
			}
		}
		checkConsistency("Post stability 2... ");

		// global satisfaction
		IntVarArgs x;
		x << insRankCourse;
		x << insRankInstructor;
		linear(*this, x, IRT_EQ, globalSat);

		//Infeasible pairs:
		for (int i = 0; i < ni; i++) {
			int i_sem =
					instance->instructors[instance->fromVirtual2RealInstr(i)].semester;
			for (int c = 0; c < nc; c++) {
				int c_sem =
						instance->courses[instance->fromVirtual2RealCrs(c)].semester;
				if (i_sem <= c_sem)
					rel(*this, course[i] != c);
			}
		}
		checkConsistency("Post infeasible pairs... ");

		// Lower bounds + symmetry breaking
		if (dynamic_cast<const FlatInstance*> (instance)!=NULL) {
			int w = 0;
			for (int j = 0; j < instance->realNi(); j++) {
				IntVarArgs tmp;
				for (int u = 0; u < instance->instructors[j].ub; u++) {
					tmp << course[w];
					w++;
				}
				if (instance->instructors[j].lb > 0)
					count(*this, tmp, nc, IRT_LQ,
							instance->instructors[j].ub
									- instance->instructors[j].lb);
				rel(*this, tmp, IRT_LQ);
			}
			checkConsistency("Post lower bounds and symmetry breaking... ");
		}
		// Minimize numb unassigned:
		count(*this, instructor, ni, IRT_EQ, nr_unassigned, opt.icl());
		//rel(*this, nr_unassigned == 0);
		checkConsistency("Post nr_unassigned... ");

		branch(*this, insRankInstructor, INT_VAR_SIZE_MIN(), INT_VAL_MIN());
		branch(*this, insRankCourse, INT_VAR_SIZE_MIN(), INT_VAL_MIN());
		branch(*this, instructor, INT_VAR_SIZE_MIN(), INT_VAL_MIN());
		branch(*this, course, INT_VAR_SIZE_MIN(), INT_VAL_MIN());
		branch(*this, nr_unassigned, INT_VAL_MIN());
		branch(*this, worstPref, INT_VAL_MIN());

	}

	virtual IntVar cost(void) const {
		//cout << "cost " << globalSat.val() << std::endl;
		//return globalSat;
		//return worstPref;
		return nr_unassigned;
	}

	/*	virtual void constrain(const Space& _b) {
	 const CIA& b = static_cast<const CIA&>(_b);

	 rel(*this, nr_assigned, IRT_LQ, b.nr_assigned.val());
	 }*/

	// Constructor for cloning s
	CIA(bool share, CIA& s) :
			IntMinimizeScript(share, s), instance(s.instance), ni(s.ni), nc(
					s.nc) {
		instructor.update(*this, share, s.instructor);
		course.update(*this, share, s.course);
		nr_unassigned.update(*this, share, s.nr_unassigned);
		worstPref.update(*this, share, s.worstPref);
		globalSat.update(*this, share, s.globalSat);
	}

	// Copy during cloning
	virtual Space*
	copy(bool share) {
		return new CIA(share, *this);
	}

	// Print solution
	virtual void print(std::ostream& os) const {
		try {
			os << "nr_unassinged = ";
			os << "(";
			for (IntVarValues i(nr_unassigned); i(); ++i)
				os << i.val() << ',';
			os << "\b";
			os << ")\n";
			os << "worstPref = ";
			os << "(";
			for (IntVarValues i(worstPref); i(); ++i)
				os << i.val() << ',';
			os << "\b";
			os << ")\n";
			os << "globalSat = ";
			os << "(";
			for (IntVarRanges i(globalSat); i(); ++i)
				os << i.min() << ".." << i.max() << " ";
			os << "\b";
			os << ")\n";
			os << "instructor = [";
			for (int j = 0; j < nc; j++) {
				os << "(";
				for (IntVarValues i(instructor[j]); i(); ++i)
					os << i.val() << ',';
				os << "\b";
				os << "),";
				//os << instructor[i].val() << ",";
			}
			os << "\b";
			os << "]\n";
			os << "course = [";
			for (int j = 0; j < ni; j++) {
				os << "(";
				for (IntVarValues i(course[j]); i(); ++i)
					os << i.val() << ',';
				os << "\b";
				os << "),";
				//os << course[i].val() << ",";
			}
			os << "\b";
			os << "]\n";

			if (true) {
				for (int j = 0; j < nc; j++) {
					os
							<< instance->courses[instance->fromVirtual2RealCrs(
									j)].name << " ";
					for (IntVarValues i(instructor[j]); i(); ++i) {
						if (i.val() >= 0 && i.val() < ni) {
							os
									<< instance->instructors[instance->fromVirtual2RealInstr(
											i.val())].name;
							os << ' ';
						}
					}
					os << "\b";
					os << std::endl;
				}
			}
		} catch (Gecode::Int::ValOfUnassignedVar e) {
			std::cerr << "Trying to access unassigned value while printing: "
					<< e.what() << std::endl;
		} catch (Exception e) {
			std::cerr << "Gecode exception while printing: " << e.what()
					<< std::endl;
			exit(1);
		}
	}
	void checkConsistency(string msg) {
		cout << msg;
		int st = status();
		//print(cout);
		if (st == SS_FAILED) //if the space is failed
			cout << "Space failed" << endl;
		else if (st == SS_BRANCH) //if the space is not failed but the space has no brancher left,
			cout << "At least one branch left" << endl;
		else if (st == SS_SOLVED) //otherwise,
			cout << "Solved" << endl;
	}
}
;

int main(int argc, char* argv[]) {
	enum {
		PROVA, MYDFS, MYGIST, ROOT
	};
	MoreOptions opt("CIA");
// Change this:
	opt.solutions(0);
	opt.time(6000);
	opt.model(PROVA);
	opt.model(PROVA, "BAB", "run via script");
	opt.model(ROOT, "Root", "Consistency at root node");
	opt.model(MYDFS, "DFS");
	opt.model(MYGIST, "GIST");
	//opt.icl(ICL_DOM);
	opt.parse(argc, argv);
//CIA cia(opt);
	cout << "Gecode version (x.y.z -> 100000x + 100y + z): "
			<< GECODE_VERSION_NUMBER << endl;

	try {
		if (opt.model() == PROVA) {
			IntMinimizeScript::run<CIA, BAB, MoreOptions> (opt);
		} else if (opt.model() == ROOT) {
			CIA* m = new CIA(opt);
			m->checkConsistency("From main... ");
		} else if (opt.model() == MYDFS) {
			CIA* m = new CIA(opt);
			int st = m->status();
			m->print(cout);
			DFS<CIA> e(m);
			delete m;
			while (CIA * s = e.next()) {
				s->print(cout);
				delete s;
			}
		} else if (opt.model() == MYGIST) {
			CIA* m = new CIA(opt);
			Gist::Print<CIA> p("Print solution");
			Gist::Options o;
			o.inspect.click(&p);
			Gist::dfs(m, o);
			delete m;
		}
	} catch (Exception e) {
		std::cerr << "Gecode exception: " << e.what() << std::endl;
		return 1;
	}

	return 0;

}

